------------------------------------------------------------------------
-- PROJECT      :   MONSOON
-- BOARD        :   MASTER CONTROL BOARD
-- DEVICE ID    :   MCB PIXEL FPGA (U20)
-- PARENT MODULE:   TOP LEVEL  
-- MODULE       :   SERIAL EEPROM READ / WRITE LOGIC
-- VERSION      :   4.0 
------------------------------------------------------------------------
-- FILE NAME    :   SERCONFIGIFC.VHD
-- AUTHOR       :   PETER MOORE
-- DATE         :   04/14/2004
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 5.2
------------------------------------------------------------------------
-- DESCRIPTION  :   THIS MODULE IMPLEMENTS A SIMPLE 12C INTERFACE TO ENABLE
-- IT TO READ AND WRITE TO A SERIAL EEPROM. FOR MONSOON, THIS EEPROM STORES
-- SPECIFIC SYSTEM ATTRIBUTES THAT THE PAN USES TO BOOT THE SYSTEM 
-- CONFIGURATION
--
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
-- THE LOGIC IS INITIALLY TRIGGERED BY THE RESET SIGNAL HOWEVER, PROVISION
-- IS MADE TO ALLOW AN EXTERNAL COMMAND TO INITIATE A READ SEQUENCE. THE
-- EEPROM DATA IS STORED IN A SHADOW RAM THAT COPIES THE CONTENTS OF THE 
-- EEPROM WHEN TRIGGERED. FOR A 24C02 DEVICE, THIS CORRESPONDS TO 64X32 BIT
-- WORDS. DURING THE RESET TRIGGER THE LOGIC SEARCHES FOR AN EEPROM DEVICE
-- THAT ASSERTS THE ACKNOWLEDGE BIT. THE SEARCH IS CONDUCTED BY FIRST LOOKING
-- FOR A DEVICE WITH ADDRESS 000 (E0,E1,E2 = '0'). SINCE THE MCB RESIDENT DEVICE
-- HAS AN ADDRESS OF 1, ANY DEVICE ON A TRANSITION BOARD WITH ADDRESS 0 WILL BE
-- READ FIRST. BY ASSERTING THE FETCH SIGNAL, ANY ADDRESS CAN BE ACCESSED VIA THE
-- LEAST SIGNIFICANT 3 BITS OF THE DATA_ADDR VECTOR. USING THE FETCH COMMAND WILL
-- OVER WRITE ANY OTHER DATA IN TEH SHADOW RAM. ONLY 63 LOCATIONS OF THE SHADOW 
-- RAM CAN BE ACCESSED. THE LAST ADDRESS (0x3F) CONTAINS THE MODULEs STATUS WORD.
-- STATUS WORD SIGNIFICANCE IS:
-- BIT  0   LSB OF PHYSICAL DEVICE      BIT 16
-- BIT  1          PHYSICAL DEVICE      BIT 17
-- BIT  2   MSB OF PHYSICAL DEVICE      BIT 18
-- BIT  3   RAM DATA VALID              BIT 19
-- BIT  4   LSB OF SCL PHASE            BIT 20
-- BIT  5   MSB OF SCL PHASE            BIT 21
-- BIT  6   LSB OF BYTE COUNTER         BIT 22
-- BIT  7   MSB OF BYTE COUNTER         BIT 23
-- BIT  8   LSB OF SHADOW RAM ADDR      BIT 24
-- BIT  9          SHADOW RAM ADDR      BIT 25
-- BIT 10          SHADOW RAM ADDR      BIT 26  WRITE SDA ENABLE FLAG
-- BIT 11          SHADOW RAM ADDR      BIT 27  READ SDA ENABLE FLAG
-- BIT 12          SHADOW RAM ADDR      BIT 28
-- BIT 13   MSB OF SHADOW RAM ADDR      BIT 29  READ PROCESS ENABLE FLAG
-- BIT 14                               BIT 30  CLOCK PROCESS ENABLE FLAG
-- BIT 15                               BIT 31  EEPROM READ IN PROGRESS

------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	       AUTHOR   VERSION  MODIFICATIONS
-- 04/14/2004   PCM    4.0  INITIAL VERSION 
------------------------------------------------------------------------
-- PORT DEFINITIONS :
-- SYSCLK   : SYSTEM CLOCK
-- RESET    : RESET THE SHADOW RAM SIGNALS AND INITIATE A SEARCH FOR EEPROMs
-- FETCH    : TRIGGERS A READ OF AN EEPROM WHOS ADDRESS IS THE SAME AS THE DATA_ADDR VALUE
-- SRCFG_CS : REQUEST TO READ THE SHADOW RAM WITH ADDRESS IN DATA_ADDR
-- DATA_OUT : DATA FROM SHADOW RAM TO SYSTRAN VIA MULTIPLEXOR
-- DATA_ADDR: ADDRESS OF DATA REQUIRED OR ADDRESS OF EEPROM DURING A FETCH CMD
-- DATA_RDY : SIGNAL TO THE SYSTRAN MULTIPLEXOR FOR DATA XFER TO PAN
-- READ     : ACKNOWLEDGE FROM SYSTRAN DATA MULTIPLEXOR
-- SCL      : I2C SERIAL CLOCK LINE
-- SDA      : I2C BIDIRECTIONAL SERIAL DATA SIGNAL
------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity SerConfigIfcV40 is
    Port ( SYSCLK   : in  std_logic;
           RESET    : in  std_logic;
           READ     : in  std_logic;
           FETCH_CS : in  std_logic;
           SRCFG_CS : in  std_logic;
           DATA_IN  : in  std_logic_vector(15 downto 0);
           DATA_OUT : inout std_logic_vector(15 downto 0);
           DATA_ADDR: in  std_logic_vector(6 downto 0);
           SCL      : out std_logic;
           SDA_IN   : in std_logic;
           SDA_OUT  : out std_logic;
           SDA_ENAB : out std_logic;
           PING     : out std_logic;
           PONG     : out std_logic
          );
end SerConfigIfcV40;

architecture Behavioral of SerConfigIfcV40 is
-- Constants used in the logic here
constant MAX_SRCH_ADDR : std_logic_vector(2 downto 0) := "100";     -- Maximum physical address that read routine will search for after reset.
constant CLK_PERIOD    : std_logic_vector(5 downto 0) := "101000";  -- Serial clock phase period in terms of sysclk. Nominally 40 (1000 ns).
constant RD_STROBE     : std_logic_vector(5 downto 0) := "010100";  -- When device data is readable.
constant TYPE_IDENT    : std_logic_vector(3 downto 0) := "1010";    -- Fixed device identity nibble.
constant BITS_PER_BYTE : std_logic_vector(3 downto 0) := "1000";    -- Number of bits in a byte ... Really !
constant MAX_RAM_ADDR  : std_logic_vector(6 downto 0) := "1111111";  -- Maximum ram address to shadow from EEPROM.
-- Shadow ram store signals
    type SerCfgStoreType is array (127 downto 0) of std_logic_vector (15 downto 0);
    signal SerCfgStore : SerCfgStoreType;               -- Actual shadow memory for serial EEPROM.
    signal SerCfgRamAddr   : std_logic_vector(6 downto 0);
    signal SerCfgStoreData : std_logic_vector(15 downto 0);
    signal SerCfgDataSrc   : std_logic_vector(15 downto 0);   
    signal SerCfgStoreWrt  : std_logic;
    signal SerCfgWrtStb    : std_logic;
    signal loc_data    : std_logic_vector(15 downto 0);
    signal Status_read : std_logic;
    signal RdEnable    : std_logic;                    -- Read decode from PAN request
    signal DevPhyAddr  : std_logic_vector(2 downto 0); -- Stores the device physical address during a ext triggered scan.

-- SerCfgReadRom process signals
type SerCfgStateType is (SerCfgIdle, SerCfgStart, SerCfgSendSlct, SerCfgSlctAck, SerCfgWait1, SerCfgSendAddr, SerCfgAddrAck, SerCfgReStart, SerCfgRdDat1, SerCfgRdDat2, SerCfgSendStop, SerCfgStop);
    signal SerCfgState : SerCfgStateType;
    signal RdRomActive  : std_logic;                    -- Signals that FSM for reading is active, Locks out the pan read during a serial read of the eeprom.
    signal ShadowAddr   : std_logic_vector(6 downto 0); -- Address of shadow ram when reading / writing the EEPROM.
    signal SerCfgDev    : std_logic_vector(2 downto 0); -- Physical address of EEPROM Address
    signal RdRomSuccess : std_logic;                    -- Indicates a successful EEPROM read (i.e. Shadow RAM data is valid).
    signal CfgReadStatus: std_logic_vector(15 downto 0); -- Status of module read back at last shadow ram address.
    signal RdLoadData   : std_logic_vector(7 downto 0); -- Data to be sent out on the SDA line
    signal WrtSdaEnable : std_logic;                    -- Enables / Disables the tristate buffer for SDA signal and to read the acknowledge reply
    signal ReStart      : std_logic;                    -- Tells the FSM that we're on our second time through the initialization phase.
    signal AckFlag      : std_logic;                    -- Used to control end of EEPROM data acquisition via read acknowledge flag
    signal ByteCntr     : std_logic_vector(1 downto 0); -- Counts the bytes from the read process. 4 bytes / shadow ram word.
    signal FirstFlag    : std_logic;                    -- Locks out address increment on shadow ram addr untill first word is in the pot.
-- SerCfgSearch process signals
    signal SerCfgRd     : std_logic;                    -- Actual trigger to start read process.
    signal lockout_read : std_logic := '0';                    -- Locks out the IC2 read routine in case reset signal is permanant. 
    signal loc_active   : std_logic;                    -- Used to cycle through actions in the search.
    signal SrchAddr     : std_logic_vector(2 downto 0); -- Physical EEPROM address used during search routine
    signal SrchTrigRdRom: std_logic;                    -- Triggers a read of the ROM in search mode.
-- SerCfgClkGen process signals
    signal ClkEnableFlag: std_logic;                    -- Enables the clock generator to run.
    signal ClkCntr      : std_logic_vector(5 downto 0); -- Counts sysclk periods to derive the I2C serial clock.
    signal SerCfgClk    : std_logic_vector(1 downto 0); -- Generates three clock phases .. 00, 01, 10 Bit 0 is inverted and used as the SCL clock. 
    signal RdStrobe     : std_logic;                    -- Ocurres in middle of clock phase 01 when SDA data should be read.
    signal ClkPhase0, ClkPhase1, ClkPhase2 : std_logic; -- Decoded clock phase information.
    signal StopLock     : std_logic;                    -- Used to block the clock to generate stop sequence
-- SerCfgWrtData process signals
    signal WrtLoadFlag  : std_logic;                    -- Request to load that data to the serial to parallel converter.
    signal SerWrtData   : std_logic_vector(7 downto 0); -- Parallel to serial conversion register.
    signal DataArm      : std_logic;                    -- One shot toggle to slow things down
    signal BitCntr      : std_logic_vector(3 downto 0); -- Serial SDA bit counter
    signal NineBits     : std_logic;                    -- Flags the end of writing out eight bits, start of bit nine time.
    signal SdaEnable : std_logic;                    -- Enables the tristate buffer for SDA signal
-- SerCfgRdData process signals
    signal SerRdData   : std_logic_vector(7 downto 0);  -- Actual serial to parallel conversion register.
    signal RdByteValue : std_logic_vector(7 downto 0);  -- Latched value of byte read in by process. Valid with RdDataRdy.
    signal RdBitCntr   : std_logic_vector(3 downto 0);  -- Counts the number of bits read in.
    signal RdEnableFlag: std_logic;                     -- Lets the read process run and collect data
    signal RdDataRdy   : std_logic;                     -- Flag to indicate that data is ready to be played with
    signal ShiftFlag   : std_logic;                     -- Used to control serial > parallel shift register.
    signal RdSdaEnable : std_logic;                     -- Enables the tristate buffer for SDA signal
    signal RdAckActive : std_logic;                     -- Used during the read acknowledge cycle.
begin

-- Debug --
    PING <= SdaEnable;
    PONG <= RdStrobe;

------------------------------------------------------------------------
-- Shadow RAM address multiplexor.
------------------------------------------------------------------------
    SerCfgRamAddr <= ShadowAddr when (RdRomActive = '1') else DATA_ADDR;
    Status_read   <= DATA_ADDR(0) and DATA_ADDR(1) and DATA_ADDR(2) and DATA_ADDR(3) and DATA_ADDR(4) and DATA_ADDR(5) and DATA_ADDR(6);
    SerCfgDataSrc <= SerCfgStoreData when (RdRomActive = '1') else DATA_IN;
    SerCfgWrtStb <= SerCfgStoreWrt when (RdRomActive = '1') else (SRCFG_CS and not READ);

------------------------------------------------------------------------
-- Shadow RAM read and write processes.
------------------------------------------------------------------------
    SerCfgWrite : process (SYSCLK)
    begin
        if (rising_edge(SYSCLK)) then
            if (SerCfgWrtStb = '1') then
                SerCfgStore(conv_integer(SerCfgRamAddr)) <= SerCfgDataSrc;
            end if;
        end if;
    end process;

    SerCfgTrig : process(SYSCLK)
    begin
        if (rising_edge(SYSCLK)) then
            if (FETCH_CS = '1' and READ = '0') then
                DevPhyAddr <= DATA_IN(2 downto 0);
            end if;
        end if;
    end process;

    loc_data <= SerCfgStore(conv_integer(SerCfgRamAddr));

------------------------------------------------------------------------
-- Shadow RAM data multiplexor.
------------------------------------------------------------------------
    RdEnable <= SRCFG_CS and READ;    
    DATA_OUT <= CfgReadStatus when (Status_read = '1' and RdEnable = '1') else (others => 'Z');
    DATA_OUT <= x"FFFF" when (RdRomActive = '1' and RdEnable = '1' and Status_read = '0') else (others => 'Z'); 
    DATA_OUT <= loc_data when (RdRomActive = '0' and RdEnable = '1' and Status_read = '0') else (others => 'Z');
    DATA_OUT <= (x"000" & '0' & DevPhyAddr) when (FETCH_CS = '1' and READ = '1') else (others => 'Z');                    

    CfgReadStatus <= RdRomActive & RdRomSuccess & ClkEnableFlag & RdEnableFlag & '0' & ShadowAddr & '0' & SerCfgDev;

------------------------------------------------------------------------
-- Initiate a search for devices on reset.
------------------------------------------------------------------------
    SerCfgSearch : process (SYSCLK)
    begin
        if(rising_edge(SYSCLK)) then
            if(RESET = '1') then
                if(lockout_read = '0') then
                    lockout_read <= '1';                -- Lock ourselves out from passing this way again until another reset.
                    SrchAddr <= (others => '0');        -- Set the search start physical address 0
                    SrchTrigRdRom <= '1';               -- Trigger the process
                    loc_active <= '0';                  -- Lock out any other action until ReadRom starts
                elsif(RdRomActive = '1') then
                    SrchTrigRdRom <= '0';               -- The read process is active so kill the trigger.
                    loc_active <= '1';
                elsif(RdRomSuccess = '0' and loc_active = '1') then
                    if (SrchAddr < MAX_SRCH_ADDR) then
                        SrchAddr <= SrchAddr + 1;       -- Increment the device address
                        SrchTrigRdRom <= '1';           -- and trigger another search.
                        loc_active <= '0';              -- Lock out this condition until ReadRom starts again.
                    end if;                                
                end if;
            else
                lockout_read <= '0';
            end if;
        end if;   
    end process;

------------------------------------------------------------------------
-- EEPROM physical address multiplexor and grouped read trigger mechanism.
------------------------------------------------------------------------

    SerCfgTrigRd : process (SYSCLK)
    begin
        if (rising_edge(SYSCLK)) then
            if (RdRomActive = '0') then                 -- We're currently in idle state
                if (SrchTrigRdRom = '1') then           -- We have work to do
                    SerCfgDev <= SrchAddr;
                    SerCfgRd  <= '1';
                elsif (FETCH_CS = '1' and READ = '0') then
                    SerCfgDev <= DATA_IN(2 downto 0);   -- Catch the relevent physical device address.
                    SerCfgRd  <= '1';                   -- And trigger the readout of the EEPROM.
                end if;
            else
                SerCfgRd <= '0';                        -- This trigger has done it's job so let it rest.
            end if;
        end if;
    end process;



------------------------------------------------------------------------
-- FSM to read out the EEPROM.
------------------------------------------------------------------------
    SerCfgReadRom : process (SYSCLK)
    begin
        if (rising_edge(SYSCLK)) then
            if(RESET = '1' and lockout_read = '0') then     -- First clock after reset true
                WrtSdaEnable <= '1';
                RdLoadData <= (others => '1');
                WrtLoadFlag <= '1';
                ClkEnableFlag <= '0';                       -- Kill the clocks
                SerCfgState <= SerCfgIdle;
            else
                case SerCfgState is
                    when SerCfgIdle =>      -- LOOKING FOR A CHANCE TO DO SOMETHING
                        WrtSdaEnable <= '1';                    -- Enable the output as SDA master.
                        WrtLoadFlag <= '0';                     -- Release the write strobe if active.
                        if (SerCfgRd = '1') then
                            RdRomActive <= '1';
                            ClkEnableFlag <= '1';               -- Start the clocks
                            RdRomSuccess <= '0';
                            RdLoadData <= (others => '0');      -- Send the start bit by requesting 
                            WrtLoadFlag <= '1';                 -- that the serial data be loaded.
                            ReStart <= '0';                     -- Kill this flag for first time through.
                            SerCfgState <= SerCfgStart;
                        else
                            RdRomActive <= '0';                 -- Go show the world we're idle.
                            StopLock <= '0';                    -- Let others use the SCL line.
                            ClkEnableFlag <= '0';               -- Kill the clocks
                            SerCfgState <= SerCfgIdle;
                        end if;
                    when SerCfgStart =>     -- WAITING FOR START SEQUENCE TO FINISH
                        WrtLoadFlag <= '0';                     -- Kill the load request.
                        if (ClkPhase0 = '1') then               -- We've got the start sequence correct
                            RdLoadData <= TYPE_IDENT & SerCfgDev & ReStart; -- Load the select code in first.
                            WrtLoadFlag <= '1';
                            SerCfgState <= SerCfgSendSlct;
                        else
                            SerCfgState <= SerCfgStart;
                        end if;
                    when SerCfgSendSlct =>  -- SEND OUT THE DEVICE SELECT BYTE
                        WrtLoadFlag <= '0';                     -- Kill the load request.
                        if (NineBits = '1') then
                            WrtSdaEnable <= '0';                -- Disable the master SDA sig and listen for an acknowledge
                            SerCfgState <= SerCfgSlctAck;
                        else
                            SerCfgState <= SerCfgSendSlct;
                        end if;
                    when SerCfgSlctAck =>   -- SEE IF THE DEVICE ACKNOWLEDGES US
                        if (RdStrobe = '1') then                -- Wait till it's time to read the SDA data
                            if (SDA_IN = '0') then                 -- Have we received an acknowledge ?
                                if (ReStart = '1') then         -- Are we currently on our second pass though ?
                                    ShadowAddr <= (others => '0'); -- Start storing data at address 0.
                                    ByteCntr   <= (others => '0'); -- and a fresh start on the byte count
                                    RdEnableFlag <= '1';           -- Enable the read FSM to run.
                                    AckFlag <= '1';                -- Enable our acknowledge strobes for continous read.
                                    FirstFlag <= '1';           -- disable address increments until after we have the first word captured.
                                    SerCfgState <= SerCfgRdDat1;-- and ready to finally read data in
                                else
                                    SerCfgState <= SerCfgWait1; -- Or we're here for the first time and need to go through restart.
                                end if;
                            else
                                RdRomSuccess <= '0';            -- Kill any pretention of success
                                SerCfgState <= SerCfgSendStop;
                            end if;
                        else
                            SerCfgState <= SerCfgSlctAck;
                        end if;
                    when SerCfgWait1 =>     -- WAIT TO RETAKE CONTROL OF THE SDA SIGNAL.
                        if (ClkPhase0 = '1') then
                            WrtSdaEnable <= '1';                 -- Re-enable the master SDA sig.
                            RdLoadData <= (others => '0');       -- Send the address we want by requesting 
                            WrtLoadFlag <= '1';                  -- that the serial data be loaded.
                            SerCfgState <= SerCfgSendAddr;
                        else
                            SerCfgState <= SerCfgWait1;
                        end if;
                    when SerCfgSendAddr =>  -- WAIT TO SEND OUT THE ADDRESS TO ACCESS IN EEPROM.
                        if( WrtLoadFlag = '1') then             -- First clock cycle at this state
                            WrtLoadFlag <= '0';                 -- So wait for the output to ratioalize and kill the load request meantime.
                            SerCfgState <= SerCfgSendAddr;
                        elsif (NineBits = '1') then             -- Else wait for the eight bits of address to go out
                            WrtSdaEnable <= '0';                -- Disable the master SDA sig and listen for an acknowledge
                            SerCfgState <= SerCfgAddrAck;
                        else
                            SerCfgState <= SerCfgSendAddr;      -- Stay here till all is done
                        end if;
                    when SerCfgAddrAck =>   -- DO WE RECEIVE AN ACKNOWLEDGE FOR THE ADDRESS ?
                        if (RdStrobe = '1') then                -- Wait for the correct time to read SDA data.
                            if (SDA_IN = '0') then                 -- Have we received an acknowledge ?
                                SerCfgState <= SerCfgReStart;
                            else
                                SerCfgState <= SerCfgSendStop;
                            end if;
                        else
                            SerCfgState <= SerCfgAddrAck;
                        end if;
                    when SerCfgReStart =>   -- RESTART THE PROTOCOL AS A PRELUDE TO READING DATA.
                        if (ClkPhase0 = '1') then               -- Wait to take back control of the SDA line
                            WrtSdaEnable <= '1';                -- Re-enable the master SDA sig.
                            RdLoadData <= (others => '0');      -- Send the start bit by requesting 
                            WrtLoadFlag <= '1';                 -- that the serial data be loaded.
                            ReStart <= '1';                     -- We are going to do this all again so tell everyone.
                            SerCfgState <= SerCfgStart;
                        else
                            SerCfgState <= SerCfgReStart;
                        end if;
                    when SerCfgRdDat1 =>    -- NOW WE BEGIN TO READ IN DATA. SET UP THE PROCESS FOR 254 BYTES TO BE READ.
                        SerCfgStoreWrt <= '0';                      -- Reset any write strobe from previous cycle.
                        if (ShadowAddr < MAX_RAM_ADDR) then
                            if (RdDataRdy = '1') then               -- We have a word .. Stash it in the appropriate place
                                case ByteCntr is
                                    when "00" =>
                                        SerCfgStoreData(7 downto 0) <= RdByteValue;
                                        if (FirstFlag = '0') then
                                            ShadowAddr <= ShadowAddr + 1;
                                        end if;    
                                    when "01" =>
                                        SerCfgStoreData(15 downto 8) <= RdByteValue;
                                        SerCfgStoreWrt <= '1';      -- set the write strobe to write to shadow ram.    
                                        FirstFlag <= '0';           -- allow address increments after we have the first word captured.
                                    when others => null;
                                end case;
                                ByteCntr <= ByteCntr + 1;           -- Increment this counter for next time
                            end if;
                            SerCfgState <= SerCfgRdDat1;
                        else
                            AckFlag <= '0';                         -- No acknowledge on this last byte.
                            SerCfgState <= SerCfgRdDat2;
                        end if;
                    when SerCfgRdDat2 =>      -- CLEAN UP THE READ PROCESS IN A GOOD WAY
                        if (RdDataRdy = '1') then                   -- finish reading the last byte to be read with no acknowledge sent
                            RdEnableFlag <= '0';                    -- Shut down the read process.
                            RdRomSuccess <= '1';                    -- Set the shadow ram data valid bit.
                            RdLoadData <= (others => '0');          -- Send the stop bit by requesting 
                            WrtSdaEnable <= '1';                    -- Enable the SDA tristate as master.
                            WrtLoadFlag <= '1';                     -- that the serial data be loaded.
                            SerCfgState <= SerCfgSendStop;
                        else
                            SerCfgState <= SerCfgRdDat2;
                        end if;                
                    when SerCfgSendStop =>  -- WAIT FOR APPROPRIATE MOMENT TO STOP
                        if (ClkPhase0 = '1') then
                           RdLoadData <= (others => '0');          -- Send the stop bit by requesting
                           StopLock <= '0';                        -- Get the phasing correct 
                           WrtLoadFlag <= '1';
                           WrtSdaEnable <= '1';                    -- Enable the SDA tristate as master to send stop bit.
                           SerCfgState <= SerCfgStop;
                        else
                            SerCfgState <= SerCfgSendStop;
                        end if;     
                    when SerCfgStop =>      -- FINISH THE JOB BY SENDING STOP SEQUENCE 
                        WrtLoadFlag <= '0';                         -- Cancel the load flag.
                        if (ClkPhase1 = '1') then                   -- Clock phase is high .. keep it there.
                            StopLock <= '1';                        -- By oring the lock signal.
                            SerCfgState <= SerCfgStop;              -- Stay in the loop.
                        elsif (ClkPhase2 = '1' and StopLock = '1') then
                            ClkEnableFlag <= '0';               -- Kill the clocks
                            RdLoadData <= (others => '1');      -- Send the stop bit now after the clock has been high for > 600ns. 
                            WrtLoadFlag <= '1';                 -- Load it.
                            SerCfgState <= SerCfgIdle;          -- Go do it all again ... someday !    
                        else
                            SerCfgState <= SerCfgStop;
                        end if;
                                  
                    when others =>
                        SerCfgState <= SerCfgIdle;    
                end case;
            end if;
        end if;
    end process;

------------------------------------------------------------------------
-- Clock generator for read / write FSMs
------------------------------------------------------------------------
    SerCfgClkGen : process (SYSCLK)
    begin
        if (rising_edge(SYSCLK)) then
            if (ClkEnableFlag = '1') then
                if(ClkCntr = CLK_PERIOD) then
                    ClkCntr <= (others => '0');         -- Time we changed the clock state
                    if (SerCfgClk(1) = '1') then        -- test the clock phase for end of cycle
                        SerCfgClk <= (others => '0');
                    else
                        SerCfgClk <= SerCfgClk + 1;     -- Generate the clock triplet
                    end if;
                else
                    if (ClkPhase1 = '1' and ClkCntr = RD_STROBE) then
                        RdStrobe <= '1';
                    else
                        RdStrobe <= '0';
                    end if;
                    ClkCntr <= ClkCntr + 1;
                end if;
            else
                ClkCntr <= (others => '0');             -- We're initially sync'ed to the start of any FSM through RdRomActive.
                SerCfgClk <= "01";                      -- Initial clock state is high.
                RdStrobe <= '0';                        -- Cancell any hung read strobe ... Murphy lurks here.
            end if;
        end if;
    end process;

    SCL <= SerCfgClk(0) or StopLock or not ClkEnableFlag; -- Output the clock state.

-- Clock phase decode.
    ClkPhase0 <=  not (SerCfgClk(0) or SerCfgClk(1));    -- Clock state low 
    ClkPhase1 <= (not SerCfgClk(1)) and SerCfgClk(0);    -- Clock state high
    ClkPhase2 <= (not SerCfgClk(0)) and SerCfgClk(1);    -- Clock state low

------------------------------------------------------------------------
-- Timing management for parallel to serial converter to SDA
------------------------------------------------------------------------
    SerCfgWrtData : process (SYSCLK)
    begin
        if (rising_edge(SYSCLK)) then
            if (RESET = '1') then
               SerWrtData <= (others => '1');      -- TAG PCM 20090626
            elsif(WrtLoadFlag = '1') then
                SerWrtData <= RdLoadData;
                BitCntr <= (others => '0');
                DataArm <= '0';
            else
                case SerCfgClk is
                    when "00" =>
                        if (DataArm = '1') then
                            SerWrtData <= SerWrtData(6 downto 0) & '1';
                            BitCntr <= BitCntr + 1;
                            DataArm <= '0';
                        end if;
                    when "01" => 
                        null;
                    when "10" =>
                        DataArm <= '1';
                    when others =>
                        null;
                end case;
            end if;
        end if;
    end process;

    SdaEnable <= WrtSdaEnable or RdSdaEnable;
    SDA_ENAB <= SdaEnable;
    SDA_OUT  <= not AckFlag when (RdSdaEnable = '1') else SerWrtData(7); -- when (SdaEnable = '1') else 'Z';  -- Output the serial data to the world.
    NineBits <= BitCntr(3);

------------------------------------------------------------------------
-- Timing management for serial to parallel converter from SDA
------------------------------------------------------------------------    
    SerCfgRdData : process (SYSCLK)
    begin
        if (rising_edge(SYSCLK)) then
            if(RdEnableFlag = '1') then
                if (RdBitCntr < BITS_PER_BYTE) then     -- HAVE WE COLLECTED ALL THE BITS YET ?
                    RdSdaEnable <= '0';                             -- Release the serial signal to the device
                    RdDataRdy <= '0';                               -- Cancel any previous data ready flag.
                    RdAckActive <= '0';                             -- and local flag used during acknowledgement.
                    if(RdStrobe = '1') then                 -- IT'S TIME TO READ THE SDA DATA BIT
                        SerRdData(0) <= SDA_IN;                     -- Catch the SDA level at the correct time
                        ShiftFlag <= '1';                           -- and prepare for a shift opperation.
                        RdBitCntr <= RdBitCntr + 1;                 -- Increment the bit counter.
                    elsif (ShiftFlag = '1') then            -- WE'VE COLLECTED, NOW SHIFT IT.
                        SerRdData <= SerRdData(6 downto 0) & '0';   -- After grabbing the bit, shift it through
                        ShiftFlag <= '0';                           -- and make sure it's only one shift.
                    end if;
                else
                    if (ClkPhase0 = '1' and RdSdaEnable = '0') then -- Okay lets start the Acknowledge to the device.
                        RdSdaEnable <= '1';                         -- Enable the tristate gate on the SDA line which sets acknowledge true
                    elsif (ClkPhase2 = '1' and RdSdaEnable = '1') then
                        RdAckActive <= '1';                         -- Wait for the 2nd clock phase to pass.
                    elsif (ClkPhase0 = '1' and RdAckActive = '1') then
                        RdSdaEnable <= '0';
                        RdByteValue <= SerRdData;                   -- Latch the data to our correspondents (redundent really, this thing is so slow !)
                        SerRdData <= (others => '0');               -- Should be redundent.
                        RdBitCntr <= (others => '0');               -- Go back and get a another byte
                        RdDataRdy <= '1';                           -- Flag our sponsor that the data is ready.
                    end if;
                end if;
            else                                        -- NOT READING ... BUT BE READY.
                RdDataRdy <= '0';                               -- Hold off the data ready flag.
                ShiftFlag <= '0';
                SerRdData <= (others => '0');                   -- Start clean ... (Hey mister clean ... Your dirty now too ! NY 2002)
                RdBitCntr <= (others => '0');
                RdSdaEnable <= '0';                             -- Allow others to use the SDA signal.
            end if;
        end if;
    end process;

end Behavioral;
